function StabData = StabilityComputation(obj,U0,Amax,Eigen_init)
% function StabData = StabilityComputation(obj,U0,Amax,Eigen_init)
% This function computes the point of changement of stability if there is a
% change of stability. This function assumes that there is only one change
% of stability of the solution, otherwise, it can give erroneous results.
%
% The exact position is computed using polynomial fits on the real parts of
% the eigen values and if this does not work, it uses a classical dichotomy
% technique.

nb_realpos1 = sum(real(Eigen_init.values)>obj.StabTol);
status_init = (nb_realpos1==0);
if status_init == 1
    status{1} = 'stable';
else
    status{1} = 'unstable';
end

Umax = evalseries(U0,Amax,obj.order);
Jacobian_str = Stab_Jacobian(obj,Umax);
[eig_val,eig_vec] = obj.Stability(Umax,Jacobian_str);
Eigen_end.values = eig_val;
Eigen_end.vectors = eig_vec;
nb_realpos2 = sum(real(eig_val)>obj.StabTol);
flag2 = (nb_realpos2==0);

Umid = Umax;
Amid = Amax-1e-12; % For the case of Amax being very close to the change of stability.

if ((nb_realpos1 ~= nb_realpos2) && (sum(sum(isnan(eig_vec)) == 0)))
    disp('Change of the stability detected. Computation of the exact point.');
    A1 = 0;
    A2 = Amax;
    nb_iter = 0;
    
    % Tolerance on relative position of Achange :
    A_tol = sqrt(obj.StabTol);
    
    try
        warning('off','MATLAB:polyfit:RepeatedPointsOrRescale');
        %% Polynomial fit - faster but sometimes does not work.
        Alist = [0 Amax];
        [~,min_ind1] = min(abs(real(Eigen_init.values)));
        [~,min_ind2] = min(abs(real(Eigen_end.values)));
        Vlist = [real(Eigen_init.values(min_ind1)) real(Eigen_end.values(min_ind2))];
        
        while (abs(Vlist(end))>obj.StabTol && (A2-A1)/A2>A_tol && nb_realpos1~=nb_realpos2 && nb_iter < obj.NRitemax)
            
            Polynom = polyfit(Alist,Vlist,nb_iter+1);
            Roots_P = roots(Polynom);
            Roots_P = Roots_P(imag(Roots_P)==0);
            Amid = Roots_P((Roots_P<=Amax & Roots_P>=0));
            
            Umid = evalseries(U0,Amid,obj.order);
            Jacobian_str=Stab_Jacobian(obj,Umid);
            
            [eig_val,eig_vec] = obj.Stability(Umid,Jacobian_str);
            nb_realpos = sum(real(eig_val)>obj.StabTol);
            
            if nb_realpos == nb_realpos1
                A1 = Amid;
                nb_realpos1 = nb_realpos;
                Alist = [Alist Amid];
                [~,min_ind] = min(abs(real(eig_val)));
                Vlist = [Vlist real(eig_val(min_ind))];
            elseif nb_realpos == nb_realpos2
                A2 = Amid;
                nb_realpos2 = nb_realpos;
                Alist = [Alist Amid];
                [~,min_ind] = min(abs(real(eig_val)));
                Vlist = [Vlist real(eig_val(min_ind))];
            else
                disp('StabilityComputation : Research did not converge.');
                nb_iter = obj.NRitemax;
            end
            nb_iter = nb_iter+1;
        end
        
        %disp(['Polynomial fit : nb_iter = ' num2str(nb_iter)]);
        
    catch
        
        A1 = 0;
        A2 = Amax;
        nb_iter = 0;
        
        %% Dichotomy - slower but always works
        while ( min(abs(real(eig_val))) && ((A2-A1)/A2>A_tol) && nb_realpos1~=nb_realpos2 && nb_iter < obj.NRitemax)
            
            Amid = (A2+A1)/2;
            Umid = evalseries(U0,Amid,obj.order);
            Jacobian_str=Stab_Jacobian(obj,Umid);
            
            [eig_val,eig_vec] = obj.Stability(Umid,Jacobian_str);
            nb_realpos = sum(real(eig_val)>obj.StabTol);
            
            if nb_realpos == nb_realpos1
                A1 = Amid;
                nb_realpos1 = nb_realpos;
            elseif nb_realpos == nb_realpos2
                A2 = Amid;
                nb_realpos2 = nb_realpos;
            else
                disp('StabilityComputation : Dichotomy did not converge.');
                nb_iter = obj.NRitemax;
            end
            nb_iter = nb_iter+1;
        end
        
        %disp(['Dichotomy : nb_iter = ' num2str(nb_iter)]);
        
    end
    
    if nb_iter == obj.NRitemax
        msgbox({'StabilityComputation : ' ; 'The location of the bifurcation is probably not accurate.'; 'Consider reducing the maximum value of Amax to compute it more accurately.'})
    end
    
    %% Post-processing
    if flag2 == 1
        status{2} = 'stable';
    else
        status{2} = 'unstable';
    end
    
    Uchange = Umid;
    Achange = Amid;
    
    bif_type = 'nothing';
    
    % Eigenvalues/Floquet exponents with positive real part
    eig1_pos = Eigen_init.values(real(Eigen_init.values)>obj.StabTol);
    eig2_pos = Eigen_end.values(real(Eigen_end.values)>obj.StabTol);
    nb_change = abs(numel(eig1_pos)-numel(eig2_pos));
    
    switch obj.type
        case {'HBQ','Q'}
            
            if nb_change == 2
                disp('Neimarck-Sacker bifurcation detected.');
                bif_type = 'NS';
            elseif nb_change == 1
                % Floquet multipliers at the bifurcation point.
                Mult = exp(2*pi/Uchange(obj.neq)*eig_val);
                prod_m1 = abs(prod(Mult-1));
                prod_p1 = abs(prod(Mult+1));
                if prod_m1 < prod_p1
                    disp('Simple bifurcation detected.');
                    bif_type = 'SB';
                else
                    disp('Period doubling bifurcation detected.');
                    bif_type = 'PD';
                end
            else
                disp('StabilityComputation : There are more than one bifurcation. Consider reducing Amax.');
                bif_type = 'nothing';
            end
        case 'AQ'
            
            if nb_change == 2
                disp('Hopf bifurcation detected.');
                bif_type = 'Hopf';
            elseif nb_change == 1
                disp('Simple bifurcation detected');
                bif_type = 'SB';
            else
                disp('StabilityComputation : There are more than one bifurcation. Consider reducing Amax.');
                bif_type = 'nothing';
            end
        otherwise
            disp('It seems that the Stability Computation of this kind of system has not been implemented yet. The results could be wrong.');
    end
    
    Eigen.type = bif_type;
    
else
    Eigen.type = 'nothing';
    Uchange = 0;
    Achange = 0;
    if status_init == 1
        status{2} = 'stable';
    else
        status{2} = 'unstable';
    end
end

[~,ind] = sort(abs(real(eig_val)));
Eigen.values = eig_val(ind);
Eigen.vectors = eig_vec(:,ind);

StabData.Eigen_init = Eigen_init;
StabData.Eigen_end = Eigen_end;
StabData.Uchange = Uchange;
StabData.Achange = Achange;
StabData.Eigen = Eigen;
StabData.status = status;

end

function [Jacobian_str] = Stab_Jacobian(obj,U)
Jacobian_str = obj.Jacobian(U);
[Jacobian_str.LK_aux,Jacobian_str.UK_aux,Jacobian_str.pK_aux,Jacobian_str.qK_aux] = lu(Jacobian_str.dRauxdUaux,'vector');
Jacobian_str.Kfull = Jacobian_str.dRdU - ...
    Jacobian_str.dRdUaux(:,Jacobian_str.qK_aux)*(Jacobian_str.UK_aux\(Jacobian_str.LK_aux\Jacobian_str.dRauxdU(Jacobian_str.pK_aux,:)));
Jacobian_str.K = Jacobian_str.Kfull(:,1:end-1);
end





%%% Using test functions only.
%     switch obj.type
%         case 'HBQ'
%             floq_mult = exp(2*pi/Umid(obj.neq)*eig_val); % exp(T*rho)
%
%             prod_m1 = abs(prod(floq_mult-1));
%             prod_p1 = abs(prod(floq_mult+1));
%             for i=1:numel(floq_mult)-1
%                 bialt_prod(i) = prod( floq_mult(i)*floq_mult(i+1:end)-1);
%             end
%             min_abs_bialt_prod = min(abs(bialt_prod));
%
%             if prod_m1 < min(prod_p1,min_abs_bialt_prod)
%                 if abs(prod(floq_mult-1)) < A_tol
%                     disp('Simple bifurcation detected.');
%                     bif_type = 'SB';
%                 end
%             elseif prod_p1 < min_abs_bialt_prod
%                 if abs(prod(floq_mult+1)) < A_tol
%                     disp('Period doubling bifurcation detected.');
%                     bif_type = 'PD';
%                 end
%             else
%                 if min_abs_bialt_prod < A_tol
%                     disp('Neimarck-Sacker bifurcation detected.');
%                     bif_type = 'NS';
%                 end
%             end
%         case 'AQ'
%
%             product_eig = abs(prod(eig_val));
%             prod_sum = Inf;
%             for i=1:numel(eig_val)-1
%                 prod_sum(i) = prod( eig_val(i) + eig_val(i+1:end));
%             end
%             min_abs_prodsum = min(abs(prod_sum));
%
%             if product_eig < min_abs_prodsum
%                 if product_eig < A_tol
%                     disp('Simple bifurcation detected');
%                     bif_type = 'SB';
%                 end
%             else
%                 if min_abs_prodsum < A_tol
%                     disp('Hopf bifurcation detected.');
%                     bif_type = 'Hopf';
%                 end
%             end
%     end
